@@ -1,3 +1,5 @@
+#preprocess
+
 /**
  * @license
  * Copyright 2015 The Emscripten Authors
@@ -13,21 +15,23 @@ var emscriptenThreadProfiler = {
 
   // Installs startup hook and periodic UI update timer.
   initialize() {
-    this.threadProfilerDiv = document.getElementById('threadprofiler');
-    if (!this.threadProfilerDiv) {
+    var self = emscriptenThreadProfiler;
+    self.threadProfilerDiv = document.getElementById('threadprofiler');
+    if (!self.threadProfilerDiv) {
       var div = document.createElement("div");
       div.innerHTML = "<div id='threadprofiler' style='margin: 20px; border: solid 1px black;'></div>";
       document.body.appendChild(div);
-      this.threadProfilerDiv = document.getElementById('threadprofiler');
+      self.threadProfilerDiv = document.getElementById('threadprofiler');
     }
-    var i = setInterval(function() { emscriptenThreadProfiler.updateUi() }, this.uiUpdateIntervalMsecs);
+    var i = setInterval(() => self.updateUi(), self.uiUpdateIntervalMsecs);
     addOnExit(() => clearInterval(i));
   },
 
   initializeNode() {
     addOnInit(() => {
-      emscriptenThreadProfiler.dumpState();
-      var i = setInterval(function() { emscriptenThreadProfiler.dumpState() }, this.uiUpdateIntervalMsecs);
+      var self = emscriptenThreadProfiler;
+      self.dumpState();
+      var i = setInterval(() => self.dumpState(), self.uiUpdateIntervalMsecs);
       addOnExit(() => clearInterval(i));
     });
   },
@@ -36,12 +40,10 @@ var emscriptenThreadProfiler = {
     var mainThread = _emscripten_main_runtime_thread_id();
 
     var threads = [mainThread];
-    for (var i in PThread.pthreads) {
-      threads.push(PThread.pthreads[i].pthread_ptr);
+    for (var thread of Object.values(PThread.pthreads)) {
+      threads.push(thread.pthread_ptr);
     }
-    for (var i = 0; i < threads.length; ++i) {
-      var threadPtr = threads[i];
-      var profilerBlock = Atomics.load(HEAPU32, (threadPtr + 8 /* {{{ C_STRUCTS.pthread.profilerBlock }}}*/) >> 2);
+    for (var threadPtr of threads) {
       var threadName = PThread.getThreadName(threadPtr);
       if (threadName) {
         threadName = `"${threadName}" (${ptrToString(threadPtr)})`;
@@ -59,17 +61,22 @@ var emscriptenThreadProfiler = {
       // initialized yet, ignore updating.
       return;
     }
+    if (!runtimeInitialized) {
+      return;
+    }
     var str = '';
     var mainThread = _emscripten_main_runtime_thread_id();
 
     var threads = [mainThread];
-    for (var i in PThread.pthreads) {
-      threads.push(PThread.pthreads[i].pthread_ptr);
+    for (var thread of Object.values(PThread.pthreads)) {
+      threads.push(thread.pthread_ptr);
     }
 
-    for (var i = 0; i < threads.length; ++i) {
-      var threadPtr = threads[i];
-      var profilerBlock = Atomics.load(HEAPU32, (threadPtr + 8 /* {{{ C_STRUCTS.pthread.profilerBlock }}}*/) >> 2);
+    for (var threadPtr of threads) {
+      var profilerBlock = Atomics.load({{{ getHeapForType('*') }}}, {{{ getHeapOffset('threadPtr + ' + C_STRUCTS.pthread.profilerBlock, '*') }}});
+#if MEMORY64
+      profilerBlock = Number(profilerBlock);
+#endif
       var threadName = PThread.getThreadName(threadPtr);
       if (threadName) {
         threadName = `"${threadName}" (${ptrToString(threadPtr)})`;
@@ -81,11 +88,11 @@ var emscriptenThreadProfiler = {
 
       var threadTimesInStatus = [];
       var totalTime = 0;
-      var offset = profilerBlock + 16/*C_STRUCTS.thread_profiler_block.timeSpentInStatus*/;
-      for (var j = 0; j < 7/*EM_THREAD_STATUS_NUMFIELDS*/; ++j, offset += 8) {
-        threadTimesInStatus.push(Number(getValue(offset, 'double')));
+      var offset = profilerBlock + {{{ C_STRUCTS.thread_profiler_block.timeSpentInStatus }}};
+      for (var j = 0; j < {{{ cDefs.EM_THREAD_STATUS_NUMFIELDS }}}; ++j, offset += 8) {
+        threadTimesInStatus.push({{{ makeGetValue('offset', 0, 'double') }}});
         totalTime += threadTimesInStatus[j];
-        setValue(offset, 0, 'double');
+        {{{ makeSetValue('offset', 0, 0, 'double') }}};
       }
       var recent = '';
       if (threadTimesInStatus[1] > 0) recent += (threadTimesInStatus[1] / totalTime * 100.0).toFixed(1) + '% running. ';
@@ -96,7 +103,7 @@ var emscriptenThreadProfiler = {
       if (recent.length > 0) str += `Recent activity: ${recent}`;
       str += '<br />';
     }
-    this.threadProfilerDiv.innerHTML = str;
+    emscriptenThreadProfiler.threadProfilerDiv.innerHTML = str;
   }
 };
 
